module(..., package.seeall)

debug = true
stereo = false

local world

--- Class Definitions
local Object = {}
local Object_mt = { __index = Object}
function Object:str()
	local s = ""
	if self.type == "camera" then
		s = "Camera"
	elseif self.type == "light" then
		s = "Light"
	else
		s = "Mesh"
	end
	s = s.." \""..self.name.."\""
	if self.parent then
		s = s.." parent \""..self.parent.name.."\""
	end
	s = s.." obj "..tostring(self.obj)
	s = s.." loc {"..self.location[1]..","..self.location[2]..","..self.location[3].."}"
	s = s.." rot {"..self.rotation[1]..","..self.rotation[2]..","..self.rotation[3].."}"
	s = s.." scale {"..self.scale[1]..","..self.scale[2]..","..self.scale[3].."}"
	
	if self.type == "camera" then
	
	elseif self.type == "light" then
		s = s.." color {"..self.color[1]..","..self.color[2]..","..self.color[3].."}"
	elseif self.type == "mesh" then
	
	end
	return s
end

--- Helper Functions
local function parent(object,p)
	if object.parent then
		for i,o in ipairs(object.parent.children) do
			if o == object then
				table.remove(object.parent.children,i)
				break
			end
		end
	end
	if p then
		E.parent_entity(object.obj,p.obj)
		table.insert(p.children,object)
	else
		E.parent_entity(object.obj,nil)
	end
	object.parent = p
end

local function splitPath(path)
	--extract the filename from the end of the path
	local a,b,dir,file = path:find('^(.-/?)([^/]-)$')
	return dir,file
end

local function getBounds(object)
	local minx,miny,minz, maxx,maxy,maxz = E.get_entity_bound(object.obj)
	local px,py,pz = E.get_entity_position(object.obj)
	return {px+object.scale[1]*minx,py+object.scale[2]*miny,pz+object.scale[3]*minz,
			object.scale[1]*(maxx-minx),object.scale[2]*(maxy-miny),object.scale[3]*(maxz-minz)}
end

local function convertHexToDec(hex)
	assert(type(hex) == "string")
	local v = 0
	for i=1,hex:len() do
		local c = hex:byte(i)
		if c >= 48 and c <= 57 then
			c = c-48
		elseif c >= 65 and c <= 70 then
			c = c-65+10
		elseif c >= 97 and c <= 102 then
			c = c-97+10
		end
		v = 16*v+c
	end
	return v
end



--- Object Interface

function loadElectroFile(path)
	world,err = loadfile(path)
	if not world then
		error("Error loading file: "..path..": "..err)
		return --just returns nil; does not cause execution to stop
	end
	world = world()
	if debug then print("Loading "..path) end
	world.path = path
	world.dir = path:sub(1,path:match("()/[^/]-$"))
	
	--now, load the various objects
	world.objectsByName,world.objectsByObj = {},{}
	if world.camera then
		if debug then print("Found Camera") end
		local obj = loadCamera(world.camera)
		if debug then print("",obj:str()) end
		world.objectsByName[obj.name] = obj
		world.objectsByObj[obj.obj] = obj
	end
	if world.lights then
		print("Found Lights")
		for i,light in ipairs(world.lights) do
			local obj = loadLight(light)
	 		if debug then print(i,obj:str()) end
			world.objectsByName[obj.name] = obj
			world.objectsByObj[obj.obj] = obj
		end
	end
	if world.meshes then
		print("Found Meshes")
		for i,mesh in ipairs(world.meshes) do
			local obj = loadMesh(mesh)
	 		if debug then print(i,obj:str()) end
			world.objectsByName[obj.name] = obj
			world.objectsByObj[obj.obj] = obj
		end
	end
end

function loadCamera(object)
	object.obj = E.create_camera(E.camera_type_perspective)
	E.set_camera_range(object.obj,.1,1000)
	object.parent = false
	if stereo then
		E.set_camera_stereo(object.obj, E.stereo_mode_tile, -.125,0,0, .125,0,0)
	end
	object.type = "camera"
	return makeObject(object)
end
local lastLight = nil
function loadLight(object)
	if not lastLight then
		lastLight = world.camera
	end
	if object.direction then
		object.obj = E.create_light(E.light_type_directional)
		E.set_entity_position(object.obj,object.direction[1],object.direction[2],object.direction[3])
	else
		object.obj = E.create_light(E.light_type_positional)
	end
	if object.color then
		E.set_light_color(object.obj,object.color[1],object.color[2],object.color[3])
	else
		object.color = {1,1,1}
	end
	object.type = "light"
	parent(object,lastLight)
	lastLight = object
	return makeObject(object)
end
local loadedMeshes = {}
function loadMesh(object)
	if not lastLight then
		lastLight = world.camera
	end
	if loadedMeshes[object.path] then
		--already loaded. Just clone the already loaded one.
		object.obj = E.create_clone(loadedMeshes[object.path].obj)
	else
		--not yet loaded. Load the mesh.
		local dir,file = splitPath(object.path)
		E.pushdir(world.dir..dir)
		object.obj = E.create_object(file)
		E.popdir()
		loadedMeshes[object.path] = object
	end
	object.type = "mesh"
	parent(object,lastLight)
	return makeObject(object)
end

function unload()
	for i,object in ipairs(world.loadedObjects) do
		E.delete_entity(object.obj)
	end
	world = nil
end

function makeObject(object)
	if object.location then
		E.set_entity_position(object.obj,object.location[1],object.location[2],object.location[3])
	else
		object.location = {0,0,0}
	end
	if object.rotation then
		E.set_entity_rotation(object.obj,object.rotation[1],object.rotation[2],object.rotation[3])
	else
		object.rotation = {0,0,0}
	end
	if object.scale then
		E.set_entity_scale(object.obj,object.scale[1],object.scale[2],object.scale[3])
	else
		object.scale = {1,1,1}
	end
	object.children = {}
	
	--apply physics
	local size = getBounds(object)
	for i=4,6 do
		if size[i] < .1 then
			size[i] = .1
		end
	end

	if object.is_body then
		E.set_entity_body_type(object.obj,true)
	else
		E.set_entity_body_type(object.obj,false)
	end
	
	if object.geom_type == "cube" or object.geom_type == "box" then
		E.set_entity_geom_type(object.obj, E.geom_type_box, size[4], size[5], size[6])
	elseif object.geom_type == "sphere" then
		E.set_entity_geom_type(object.obj, E.geom_type_sphere, math.max(size[4],size[5],size[6])/2)
	elseif object.geom_type == "ray" then
		E.set_entity_geom_type(object.obj, E.geom_type_ray, 0, 0, 0, 0, -100, 0)
	else
		E.set_entity_geom_type(object.obj, E.geom_type_none)
	end
	
	if object.category then
		if type(object.category) == "string" then
			object.category = convertHexToDec(object.category)
		end
		E.set_entity_geom_attr(object.obj, E.geom_attr_category, object.category) --all 1's
	else
		E.set_entity_geom_attr(object.obj, E.geom_attr_category, 0)
	end
	if object.collider then
		if type(object.collider) == "string" then
			object.collider = convertHexToDec(object.collider)
		end
		E.set_entity_geom_attr(object.obj, E.geom_attr_collider, object.collider) --all 1's
	else
		E.set_entity_geom_attr(object.obj, E.geom_attr_collider, 0)
	end
	if object.response then
		if type(object.response) == "string" then
			object.response = convertHexToDec(object.response)
		end
		E.set_entity_geom_attr(object.obj, E.geom_attr_response, object.response) --all 1's
	else
		E.set_entity_geom_attr(object.obj, E.geom_attr_response, 0)
	end
	if object.callback then
		if type(object.callback) == "string" then
			object.callback = convertHexToDec(object.callback)
		end
		E.set_entity_geom_attr(object.obj, E.geom_attr_callback, object.callback) --all 1's
	else
		E.set_entity_geom_attr(object.obj, E.geom_attr_callback, 0)
	end
	
	
	if object.mass then
		E.set_entity_geom_attr(object.obj, E.geom_attr_mass, tostring(object.mass))
	else
		E.set_entity_geom_attr(object.obj, E.geom_attr_mass, 0)
	end
	
	if object.has_gravity then
		E.set_entity_body_attr(object.obj, E.body_attr_gravity, true)
	else
		E.set_entity_body_attr(object.obj, E.body_attr_gravity, false)
	end
	
	--make object will add the appropriate methods to the object table
	setmetatable(object,Object_mt)

	return object
end

function getObjectByName(name)
	return world.objectsByName[name]
end
function getEntityByName(name)
	return getObjectByName(name).obj
end
function getObjectByObj(obj)
	return world.objectsByObj[obj]
end
