#!BPY

"""
Name: 'Wavefront Individiual (multiple .obj)...'
Blender: 245
Group: 'Export'
Tooltip: 'Save a Wavefront OBJ File, Individually'
"""

__author__ = "Matt Handley"
__url__ = ['www.evl.uic.edu']
__version__ = "0.2"

__bpydoc__ = """\
This script is an exporter to electro. It is a rewrite of the original export_obj.py included with Blender, and most of the obj and mtl file writing comes from there.

Usage:

Select the objects you wish to export and run this script from "File->Export" menu.
Selecting the default options from the popup box will be good in most cases.
All objects that can be represented as a mesh (mesh, curve, metaball, surface, text3d)
will be exported as mesh data.
"""


# --------------------------------------------------------------------------
# Electro Export v0.2 by Matt Handley (applmak)
# --------------------------------------------------------------------------
# ***** BEGIN GPL LICENSE BLOCK *****
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
# ***** END GPL LICENCE BLOCK *****
# --------------------------------------------------------------------------


import Blender
from Blender import Mesh, Scene, Window, sys, Image, Draw
import BPyMesh
import BPyObject
import BPySys
import BPyMessages
import math
import os

## Classes
class ElectroCamera:
	def __init__(self, object):
		global bExportFlags
		# get basic camera properties
		location = object.getLocation('worldspace')
		rotation = [r * 180/3.1415926 for r in object.getEuler('worldspace')] #convert to degrees
		properties = object.getAllProperties()
		name = object.getName()
		
		self.desc = "camera={name=%r" % name
		if not (location[0] == 0 and location[1] == 0 and location[2] == 0):
			self.desc += ", location={%.6f,%.6f,%.6f}" % tuple(location)
		if not (rotation[0] == 0 and rotation[1] == 0 and rotation[2] == 0):
			self.desc += ", rotation={%.6f,%.6f,%.6f}" % tuple(rotation)
		if bExportFlags.val:
			for p in properties:
				self.desc += ", "+p.getName()+" = \""+str(p.getData())+"\""
		self.desc += "}"
	def __str__(self):
		return self.desc

class ElectroLight:
	def __init__(self, object):
		global bExportFlags
		# get basic light properties
		location = object.getLocation('worldspace')
		properties = object.getAllProperties()
		name = object.getName()
		
		self.desc = "{name=%r" % name
		if object.getData().type == 1: # a directional light
			if not (location[0] == 0 and location[1] == 0 and location[2] == 0):
				self.desc += ", direction={%.6f,%.6f,%.6f}" % tuple(location)
		else: # all other lights (assumed to be positional)
			if not (location[0] == 0 and location[1] == 0 and location[2] == 0):
				self.desc += ", location={%.6f,%.6f,%.6f}" % tuple(location)
		
		#color
		self.desc += ", color={%.3f,%.3f,%.3f}" % tuple(object.getData().col)
		
		if bExportFlags.val:
			for p in properties:
				self.desc += ", "+p.getName()+" = \""+str(p.getData())+"\""
		self.desc += "}"
	def __str__(self):
		return self.desc

class ElectroMesh:
	def __init__(self,object):
		global bExportFlags
		# get basic mesh properties
		location = object.getLocation('worldspace')
		rotation = [180/3.1415926 * f for f in object.getEuler('worldspace')]
		rotMatrix = Blender.Mathutils.RotationMatrix(rotation[0],3,'x')*Blender.Mathutils.RotationMatrix(rotation[1],3,'y')*Blender.Mathutils.RotationMatrix(rotation[2],3,'z')
		rotation2 = [\
			180/3.1415926*-math.asin(rotMatrix[0][2]),\
			180/3.1415926*math.atan2(rotMatrix[1][2],rotMatrix[2][2]),\
			180/3.1415926*math.atan2(rotMatrix[0][1],rotMatrix[0][0])
		]
		
		scale = object.getSize('worldspace')
		properties = object.getAllProperties()
		name = object.getName()
		
		self.desc = "{name=%r" % name
		#save mesh path
		self.desc += ", path=%r" % (makeKey(object.getData(mesh=1))+".obj")
		if not (location[0] == 0 and location[1] == 0 and location[2] == 0):
			self.desc += ", location={%.6f,%.6f,%.6f}" % tuple(location)
		#if not (rotation[0] == 0 and rotation[1] == 0 and rotation[2] == 0):
		self.desc += ", oldrotation={%.6f,%.6f,%.6f}" % tuple(rotation)
		self.desc += ", rotation={%.6f,%.6f,%.6f}" % tuple(rotation2)
		if not (math.fabs(scale[0]-1) < .001 and math.fabs(scale[1]-1) < .001 and math.fabs(scale[2]-1) < .001):
			self.desc += ", scale={%.6f,%.6f,%.6f}" % tuple(scale)
		if bExportFlags.val:
			for p in properties:
				self.desc += ", "+p.getName()+" = \""+str(p.getData())+"\""
		self.desc += "}"
		
		self.object = object
		
	def __str__(self):
		return self.desc
	
	def write(self):
		global electroFilePath
		global containerMesh
		name = makeKey(self.object.getData(mesh=1))
		filename = os.path.join(os.path.dirname(electroFilePath),name+".obj")
		print("* Writing Mesh to: "+filename)
		file = open(filename,"w") #overwriting!
		
		file.write('# Blender->Electro v%s OBJ File: %s\n' % (Blender.Get('version'), os.path.basename(Blender.Get('filename')) ))
		
		if containerMesh == None:
			for meshName in Blender.NMesh.GetNames():
				if meshName.startswith(tempMeshName):
					tempMesh = Mesh.Get(meshName)
					if not tempMesh.users:
						containerMesh = tempMesh
			if not containerMesh:
				containerMesh = Mesh.New(tempMeshName)
		
		scene = Scene.GetCurrent()
		mesh = BPyMesh.getMeshFromObject(self.object, containerMesh, True, False, scene)
		if not mesh:
			#Whoa!
			raise Exception("I somehow can't get a mesh for something that I should get a mesh for!" + name)
		
		materials = mesh.materials
		# Initialize totals, these are updated each object
		totverts = totuvco = totno = 1
		face_vert_index = 1 # used for uvs now
		globalNormals = {}
		faceuv = mesh.faceUV
		
		faces = [f for f in mesh.faces]
		
		if len(faces)+len(mesh.verts) == 0:
			print("*\tNothing to do for this mesh")
			return
		
		if faces:
			BPyMesh.meshCalcNormals(mesh)
		
		materialNames = []
		if materials:
			for mat in materials:
				if mat: # !=None
					materialNames.append(mat.name)
				else:
					materialNames.append(None)
		
		# Possible there null materials, will mess up indicies
		# but at least it will export, wait until Blender gets fixed.
		materialNames.extend((16-len(materialNames)) * [None])
			
		if faceuv:
			try: faces.sort(key = lambda a: (a.mat, a.image, a.smooth))
			except: faces.sort(lambda a,b: cmp((a.mat, a.image, a.smooth), (b.mat, b.image, b.smooth)))
		elif len(materials) > 1:
			try:	faces.sort(key = lambda a: (a.mat, a.smooth))
			except:	faces.sort(lambda a,b: cmp((a.mat, a.smooth), (b.mat, b.smooth)))
		else:
			# no materials
			try:	faces.sort(key = lambda a: a.smooth)
			except:	faces.sort(lambda a,b: cmp(a.smooth, b.smooth))
		
		# Set the default mat to no material and no image.
		contextMat = (0, 0) # Can never be this, so we will label a new material teh first chance we get.
		contextSmooth = None # Will either be true or false,  set bad to force initialization switch.
		
		objectName = self.object.name
		if objectName == name:
			obnamestring = fixName(objectName)
		else:
			obnamestring = '%s_%s' % (fixName(objectName), fixName(name))
		
		file.write('o %s\n' % obnamestring) # Write Object name
		
		# Vert
		for v in mesh.verts:
			file.write('v %.6f %.6f %.6f\n' % tuple(v.co))
		
		# UV
		if faceuv:
			for f in faces:
				for uv in f.uv:
					file.write('vt %.6f %.6f 0.0\n' % tuple(uv))
		
		for f in faces:
			for v in f:
				noKey = veckey3d(v.no)
				if not globalNormals.has_key( noKey ):
					globalNormals[noKey] = totno
					totno += 1
					file.write('vn %.6f %.6f %.6f\n' % noKey)
	
		for f in faces:
			f_v = f.v
			if faceuv:
				f_uv= f.uv
				
			# MAKE KEY
			if faceuv and f.image: # Object is always true.
				key = materialNames[min(f.mat,len(materialNames)-1)],  f.image.name
			else:
				key = materialNames[min(f.mat,len(materialNames)-1)],  None # No image, use None instead.
				
			# CHECK FOR CONTEXT SWITCH
			if key == contextMat:
				pass # Context alredy switched, dont do anythoing
			else:
				if key[0] == None and key[1] == None:
					# Write a null material, since we know the context has changed.
					file.write('mtllib Default.mtl\n')
					file.write('usemtl Default\n') # mat, image
						
				else:
					if key[1] == None:
						matstring = '%s' % fixName(key[0])
					else:
						matstring = '%s_%s' % (fixName(key[0]), fixName(key[1]))
					try:
						file.write('mtllib %s.mtl\n' % makeKey(materials[f.mat]))
					except:
						file.write('mtllib Default.mtl\n')
					file.write('usemtl %s\n' % matstring) # can be mat_image or (null)
					
			contextMat = key
			
			file.write('s 1\n')
			
			file.write('f')
			if faceuv:
				for vi, v in enumerate(f_v):
					file.write(' %d/%d/%d' % (\
					  v.index+totverts,\
					  face_vert_index + vi,\
					  globalNormals[ veckey3d(v.no) ])) # vert, uv, normal
			
				face_vert_index += len(f_v)
			else: # No UV's
				for v in f_v:
					file.write(' %d//%d' % (\
					  v.index+totverts,\
					  globalNormals[ veckey3d(v.no) ]))
			file.write('\n')

## drawing callbacks
def Gui():
	global electroFilePath
	global bFilePath,bExportCamera,bExportLights,bExportMeshes,bExportSelection,bInstantiateDups,bExportFlags,bGroup,bAnimation

	Blender.BGL.glClearColor(.75,.75,.75,1)
	Blender.BGL.glClear(Blender.BGL.GL_COLOR_BUFFER_BIT)
	
	size = Blender.Window.GetAreaSize()
	
	Blender.BGL.glColor3f(.1,.1,.1)
	Blender.BGL.glRasterPos2i(10,size[1]-20)
	Blender.Draw.Text("Electro Export:","normal")

	Blender.BGL.glRasterPos2i(25,size[1]-40)
	Blender.Draw.Text("Electro File:","normal")
	bFilePath = Blender.Draw.PushButton(electroFilePath,1,100,size[1]-45,500,18,"choose electro file path",doChooseFile)

	Blender.BGL.glRasterPos2i(25,size[1]-60)
	Blender.Draw.Text("Export Only:","normal")
	try:
		bExportCamera = Blender.Draw.Toggle("Camera",0,100,size[1]-65,100,18,bExportCamera.val,"export the camera")
	except:
		bExportCamera = Blender.Draw.Toggle("Camera",0,100,size[1]-65,100,18,1,"export the camera")
	try:
		bExportLights = Blender.Draw.Toggle("Lights",0,200,size[1]-65,100,18,bExportLights.val,"export the lights")
	except:
		bExportLights = Blender.Draw.Toggle("Lights",0,200,size[1]-65,100,18,1,"export the lights")
	try:
		bExportMeshes = Blender.Draw.Toggle("Meshes",0,300,size[1]-65,100,18,bExportMeshes.val,"export the meshes")
	except:
		bExportMeshes = Blender.Draw.Toggle("Meshes",0,300,size[1]-65,100,18,1,"export the meshes")
	try:
		bExportSelection = Blender.Draw.Toggle("Selection",0,400,size[1]-65,100,18,bExportSelection.val,"only consider the objects in the selection")
	except:
		bExportSelection = Blender.Draw.Toggle("Selection",0,400,size[1]-65,100,18,0,"only consider the objects in the selection")
	
	Blender.BGL.glRasterPos2i(25,size[1]-80)
	Blender.Draw.Text("Options:","normal")
	
	'''try:
		bInstantiateDups = Blender.Draw.Toggle("Auto-instantiate",0,40,size[1]-105,100,18,bInstantiateDups.val,"auto-instantiate meshes with dupliframes and empties with dupligroups")
	except:
		bInstantiateDups = Blender.Draw.Toggle("Auto-instantiate",0,40,size[1]-105,100,18,0,"auto-instantiate meshes with dupliframes and empties with dupligroups")'''
	try:
		bExportFlags = Blender.Draw.Toggle("Electro Flags",0,140,size[1]-105,100,18,bExportFlags.val,"export all electro flags (physics)")
	except:
		bExportFlags = Blender.Draw.Toggle("Electro Flags",0,140,size[1]-105,100,18,1,"export all electro flags (physics)")
	'''try:
		bGroup = Blender.Draw.Toggle("Group",0,240,size[1]-105,100,18,bGroup.val,"export blender group as a single .obj")
	except:
		bGroup = Blender.Draw.Toggle("Group",0,240,size[1]-105,100,18,0,"export blender group as a single .obj")
	try:
		bAnimation = Blender.Draw.Toggle("As Animation",0,340,size[1]-105,100,18,bAnimation.val,"export animation (be warned. this may make a lot of files.)")
	except:
		bAnimation = Blender.Draw.Toggle("As Animation",0,340,size[1]-105,100,18,0,"export animation (be warned. this may make a lot of files.)")'''

	bExport = Blender.Draw.PushButton("Export",1,40,size[1]-130,80,18,"export",doExport)
	bCancel = Blender.Draw.PushButton("Cancel",1,125,size[1]-130,80,18,"cancel",doCancel)
def Event(evt, val):
	pass
def ButtonEvent(evt):
	pass

## Electro File Saving Callbacks
def doChooseFile(evt, val):
	Window.FileSelector(endChooseFile, 'Save', sys.makename(ext='.electro'))
def endChooseFile(filename):
	global electroFilePath
	electroFilePath = filename

## Handles the cancel button (exits without doing anything)
def doCancel(evt, val):
	Blender.Draw.Exit()

## Handles the export button
materialLibrary = {}
meshLibrary = {}
tempMeshName = "~temp_mesh"
def doExport(evt, val):
	global electroFilePath
	global bFilePath,bExportCamera,bExportLights,bExportMeshes,bExportSelection,bInstantiateDups,bExportFlags,bGroup,bAnimation
	global containerMesh
	containerMesh = None
	
	if electroFilePath == "...":
		return
	
	#turn on the wait cursor
	Blender.Window.WaitCursor(1)
	Blender.Window.EditMode(0)
	
	if not electroFilePath.lower().endswith('.electro'):
		electroFilePath += '.electro'

	print("ElectroFilePath is:"+electroFilePath)

	electroFileDirectory, electroFileName = os.path.split(electroFilePath)
	
	#save the current scene
	scene = Scene.GetCurrent()
	
	# Export all scenes.
	context = scene.getRenderingContext()
	orig_frame = Blender.Get('curframe')

	#if bAnimation.val:
	#	scene_frames = xrange(context.startFrame(), context.endFrame()+1)
	#else:
	scene_frames = [orig_frame] # Dont export an animation.
	
	print("Exporting the following frames of the current scene:"+",".join([str(i) for i in scene_frames]))
	containerMesh = Mesh.New(tempMeshName)
	# Loop through all frames in the scene and export.
	for frame in scene_frames:
		Blender.Set('curframe', frame)
		if bExportSelection.val:
			potential_objects = scene.objects.context
		else:	
			potential_objects = scene.objects
		
		print("Exporting "+str(len(potential_objects))+" from frame "+str(frame));
		
		#do the first pass
		print("\tFirst Pass:")
		#	prune from the set (camera, lights, meshes)
		export_camera = None
		export_lights = []
		export_meshes = []
		
		for someObject in potential_objects:
			for object,ob_mat in BPyObject.getDerivedObjects(someObject):
				print("\t\tConsidering "+object.name)
				if object.type == 'Camera' and bExportCamera.val:
					print("\t\t\t--> Camera")
					export_camera = object
				elif (object.type == 'Lamp' or object.type == 'Sun') and bExportLights.val:
					print("\t\t\t--> Light")
					export_lights.append(object)
				elif bExportMeshes.val:
					mesh = BPyMesh.getMeshFromObject(object, containerMesh, True, False, scene)
					if mesh: #this object CAN be transformed into a mesh. Really.
						print("\t\t\t--> Mesh")
						export_meshes.append(object)
						#	for the meshes, grab the unique materials and mesh datas. Assign them unique names.
						materials = mesh.materials
						#	somehow, we need to generate a key for this datablock
						if len(materials) == 0:
							# default material
							print("\t\t\tMaterial key: default")
						else:
							for m in materials:
								if m == None:
									continue
								key = makeKey(m)
								print("\t\t\tMaterial key: "+key)
								# add to material library
								materialLibrary[key] = m
						
						meshKey = makeKey(object.getData(mesh=1))
						print("\t\t\tMesh Key:"+meshKey)
						meshLibrary[meshKey] = object.getData(mesh=1)
					
		
		#do the second pass
		electroFileContents = "return {\n"
		print("\tSecond Pass:")
		if export_camera:
			print("\t\tWriting Camera: "+export_camera.name)
			electroFileContents += "\t"+str(ElectroCamera(export_camera)) + ",\n"
		if len(export_lights) > 0:
			electroFileContents += "\tlights={\n"
			for light in export_lights:
				print("\t\tWriting Light: "+light.name)
				electroFileContents += "\t\t"+str(ElectroLight(light)) + ",\n"
			electroFileContents += "\t},\n"

		#	write the mesh data
		if len(export_meshes) > 0:
			electroFileContents += "\tmeshes = {\n"
			meshesAlreadyWritten = {}
			for object in export_meshes:
				print("\t\tWriting "+object.name)
				electroSpecification = ElectroMesh(object)
				meshKey = makeKey(object.getData(mesh=1))
				try:
					alreadyWritten = meshesAlreadyWritten[meshKey]
				except KeyError:
					print("\t\t\tMesh:"+meshKey)
					electroSpecification.write()
					meshesAlreadyWritten[meshKey] = True
				
				# now, right here... add to electro file
				electroFileContents += "\t\t"+str(electroSpecification) + ",\n"
			electroFileContents += "\t}\n"
		electroFileContents += "}\n"
		
	#	write the mtl files
	for key, material in materialLibrary.iteritems():
		print("Writing material: "+key)
		writeMaterialFile(key,material)
	
	print("Writing Default Material.")
	writeDefaultMaterial()
	
	#actually write the electro file
	print("Writing electro file to: "+electroFilePath)
	file = open(electroFilePath,"w") #overwriting!
	file.write(electroFileContents)
	file.close()
	
	Blender.Set('curframe', orig_frame)
	
	# Restore old active scene.
	Window.WaitCursor(0)
	
	#finish up
	Blender.Draw.Exit()

## material helper functions
def writeDefaultMaterial():
	global electroFilePath
	world = Blender.World.GetCurrent()
	if world:
		worldAmb = world.getAmb()
	else:
		worldAmb = (0,0,0) # Default value
	file = open(os.path.join(os.path.dirname(electroFilePath),"Default.mtl"),"w") #overwrite
	file.write('# Blender->Electro MTL File: %s\n' % os.path.basename(Blender.Get('filename')))
	file.write('newmtl Default\n')
	file.write('Ns 0\n')
	file.write('Ka %.6f %.6f %.6f\n' %  tuple([c for c in worldAmb])  ) # Ambient, uses mirror colour,
	file.write('Kd 0.8 0.8 0.8\n')
	file.write('Ks 0.8 0.8 0.8\n')
	file.write('d 1\n') # No alpha
	file.write('illum 2\n') # light normaly
	file.close()

def writeMaterialFile(materialName,material):
	global electroFilePath
	world = Blender.World.GetCurrent()
	if world:
		worldAmb = world.getAmb()
	else:
		worldAmb = (0,0,0) # Default value
	file = open(os.path.join(os.path.dirname(electroFilePath),materialName+".mtl"),"w") #overwrite
	file.write('# Blender->Electro MTL File: %s\n' % os.path.basename(Blender.Get('filename')))
	file.write('newmtl %s\n' % materialName)
	file.write('Ns %.6f\n' % ((material.getHardness()-1)*1.9607843137254901))
	file.write('Ka %.6f %.6f %.6f\n' % tuple([c*material.amb for c in worldAmb])  )
	file.write('Kd %.6f %.6f %.6f\n' % tuple([c*material.ref for c in material.rgbCol]) )
	file.write('Ks %.6f %.6f %.6f\n' % tuple([c*material.spec for c in material.specCol]) )
	file.write('Ni %.6f\n' % material.IOR)
	file.write('d %.6f\n' % material.alpha) # No alpha
	if material.getMode() & Blender.Material.Modes['SHADELESS']:
		file.write('illum 0\n') # light normaly
	elif material.getSpec() == 0:
		file.write('illum 1\n')
	else:
		file.write('illum 2\n')
	file.close()

## Utility & Classes
# Returns a tuple - path,extension.
# 'hello.obj' >  ('hello', '.obj')
def splitExt(path):
	dotidx = path.rfind('.')
	if dotidx == -1:
		return path, ''
	else:
		return path[:dotidx], path[dotidx:] 

# Used to add the scene name into the filename without using odd chars
def fixName(name):
	if name == None:
		return 'None'
	else:
		return name.replace(' ', '_')

def copy_file(source, dest):
	file = open(source, 'rb')
	data = file.read()
	file.close()
	
	file = open(dest, 'wb')
	file.write(data)
	file.close()


def copy_images(dest_dir):
	if dest_dir[-1] != sys.sep:
		dest_dir += sys.sep
	
	# Get unique image names
	uniqueImages = {}
	for matname, imagename in MTL_DICT.iterkeys(): # Only use image name
		# Get Texface images
		if imagename != None:
			uniqueImages[imagename] = None # Should use sets here. wait until Python 2.4 is default.
		
		# Get MTex images
		if matname != None:
			mat= Material.Get(matname)
			for mtex in mat.getTextures():
				if mtex and mtex.tex.type == Blender.Texture.Types.IMAGE:
					try:
						uniqueImages[mtex.tex.image.name] = None
					except:
						pass
	
	# Now copy images
	copyCount = 0
	
	for imageName in uniqueImages.iterkeys():
		bImage = Image.Get(imageName)
		image_path = sys.expandpath(bImage.filename)
		if sys.exists(image_path):
			# Make a name for the target path.
			dest_image_path = dest_dir + image_path.split('\\')[-1].split('/')[-1]
			if not sys.exists(dest_image_path): # Image isnt alredy there
				#print '\tCopying "%s" > "%s"' % (image_path, dest_image_path)
				copy_file(image_path, dest_image_path)
				copyCount+=1
	#print '\tCopied %d images' % copyCount

def veckey3d(v):
	return round(v.x, 6), round(v.y, 6), round(v.z, 6)

def veckey2d(v):
	return round(v.x, 6), round(v.y, 6)	
	
def makeKey(thing):
	if thing.lib == None:
		return fixName(thing.name)
	else:
		libIndex = hashLib(thing.lib)
		return fixName(thing.name)+str(libIndex)

global nextLibraryIndex,libraryPathHash
libraryPathHash = {}
nextLibraryIndex = 1
def hashLib(lib):
	global nextLibraryIndex,libraryPathHash
	try:
		return libraryPathHash[lib]
	except KeyError:
		#there is no thing.lib in the path hash
		libraryPathHash[lib] = nextLibraryIndex
		nextLibraryIndex += 1
		return libraryPathHash[lib]
		
## main
if __name__ == '__main__':
	global electroFilePath
	electroFilePath = '...'
	Blender.Draw.Register(Gui, Event, ButtonEvent)

